#include <stdio.h>
#include <stdlib.h>
#include <gccore.h>
#include "wiiuse/wpad.h"

static void *xfb = NULL;
static GXRModeObj *rmode = NULL;

static const char *extension_name(int type)
{
	switch(type)
	{
		case EXP_NUNCHUK:
			return "Nunchuk";
		case EXP_CLASSIC:
			return "Classic";
		case EXP_GUITAR_HERO_3:
			return "Guitar Hero 3";
		case EXP_WII_BOARD:
			return "Wii board";
		case EXP_MOTION_PLUS:
			return "Wii motion plus";
		case EXP_EXTENMOTE_N64:
			return "Extenmote N64";
		case EXP_EXTENMOTE_GC:
			return "Extenmote Gamecube";
		case EXP_EXTENMOTE_SNES:
			return "Extenmote SNES";
		case EXP_EXTENMOTE_NES:
			return "Extenmote NES";
		case EXP_NONE:
			return "None";
		default:
			return "Unknown";
	}
}

void printHex(unsigned char *dat, int len)
{
	do {
		printf("%02X ", *dat);
		dat++;
	} while(--len);
}

void printN64Buttons(unsigned char dat[2])
{
	if (dat[0] & 0x80) printf("A ");
	if (dat[0] & 0x40) printf("B ");
	if (dat[0] & 0x20) printf("Z ");
	if (dat[0] & 0x10) printf("START ");
	if (dat[0] & 0x08) printf("Up ");
	if (dat[0] & 0x04) printf("Down ");
	if (dat[0] & 0x02) printf("Left ");
	if (dat[0] & 0x01) printf("Right ");
	if (dat[1] & 0x20) printf("L ");
	if (dat[1] & 0x10) printf("R ");
	if (dat[1] & 0x08) printf("C-Up ");
	if (dat[1] & 0x04) printf("C-Down ");
	if (dat[1] & 0x02) printf("C-Left ");
	if (dat[1] & 0x01) printf("C-Right ");
}

void printGCButtons(unsigned char dat[2])
{
	if (dat[0] & 0x10) printf("START ");
	if (dat[0] & 0x08) printf("Y ");
	if (dat[0] & 0x04) printf("X ");
	if (dat[0] & 0x02) printf("B ");
	if (dat[0] & 0x01) printf("A ");
	if (dat[1] & 0x40) printf("L ");
	if (dat[1] & 0x20) printf("R ");
	if (dat[1] & 0x10) printf("Z ");
	if (dat[1] & 0x08) printf("Up ");
	if (dat[1] & 0x04) printf("Down ");
	if (dat[1] & 0x02) printf("Right ");
	if (dat[1] & 0x01) printf("Left ");
}

void printSNESButtons(unsigned char dat[2])
{
	if (dat[0] & 0x80) printf("B ");
	if (dat[0] & 0x40) printf("Y ");
	if (dat[0] & 0x20) printf("SELECT ");
	if (dat[0] & 0x10) printf("START ");
	if (dat[0] & 0x08) printf("Up ");
	if (dat[0] & 0x04) printf("Down ");
	if (dat[0] & 0x02) printf("Left ");
	if (dat[0] & 0x01) printf("Right ");

	if (dat[1] & 0x80) printf("A ");
	if (dat[1] & 0x40) printf("X ");
	if (dat[1] & 0x20) printf("L ");
	if (dat[1] & 0x10) printf("R ");
}

void printNESButtons(unsigned char dat[1])
{
	if (dat[0] & 0x80) printf("A ");
	if (dat[0] & 0x40) printf("B ");
	if (dat[0] & 0x20) printf("SELECT ");
	if (dat[0] & 0x10) printf("START ");
	if (dat[0] & 0x08) printf("Up ");
	if (dat[0] & 0x04) printf("Down ");
	if (dat[0] & 0x02) printf("Left ");
	if (dat[0] & 0x01) printf("Right ");
}

static void printClassicData(struct classic_ctrl_t *classic)
{
	printf("\nClassic controller data {\n");
	printf("  Buttons: %04x\n", classic->btns_held);
	printf("  rs_raw: %02x\n", classic->rs_raw);
	printf("  ls_raw: %02x\n", classic->ls_raw);
	printf("  r_shoulder: %.3f\n", classic->r_shoulder);
	printf("  l_shoulder: %.3f\n", classic->l_shoulder);
	printf("  x: %02x, y: %02x\n", classic->ljs.pos.x, classic->ljs.pos.y);
	printf("  rx: %02x, ry: %02x\n", classic->rjs.pos.x, classic->rjs.pos.y);
	printf("\n");
	printf("  Is GC adapter: %d\n", classic->raw_gc.valid);
	printf("  Is N64 adapter: %d\n", classic->raw_n64.valid);
	printf("  Is SNES adapter: %d\n", classic->raw_snes.valid);
	printf("  Is NES adapter: %d\n", classic->raw_nes.valid);
	printf("\n");

	if (classic->raw_n64.valid) {
		printf("  N64 raw data { ");
		printf("    raw: ");
		printHex(classic->raw_n64.data, 4);
		printf("\n    Buttons: ");
		printN64Buttons(classic->raw_n64.data);
		printf("\n    Analog: x=0x%02x, y=0x%02x\n",
				classic->raw_n64.data[2],
				classic->raw_n64.data[3]);
		printf("  }\n");
	}

	if (classic->raw_gc.valid) {
		printf("  Gamecube raw data {\n");
		printf("    raw: ");
		printHex(classic->raw_gc.data, 4);
		printf("\n    Buttons: ");
		printGCButtons(classic->raw_gc.data);
		printf("\n    Analog: x=0x%02x, y=0x%02x, cx=0x%02x, cy=0x%02x, l=0x%02x, r=0x%02x\n",
					classic->raw_gc.data[2],
					classic->raw_gc.data[3],
					classic->raw_gc.data[4],
					classic->raw_gc.data[5],
					classic->raw_gc.data[6],
					classic->raw_gc.data[7]);
		printf("  }\n");
	}

	if (classic->raw_snes.valid) {
		printf("  SNES raw data {\n");
		printf("    raw: ");
		printHex(classic->raw_snes.data, 2);
		printSNESButtons(classic->raw_snes.data);
		printf("  }\n");
	}

	if (classic->raw_nes.valid) {
		printf("  NES raw data {\n");
		printf("    raw: ");
		printHex(classic->raw_nes.data, 1);
		printNESButtons(classic->raw_nes.data);
		printf("  }\n");
	}

	printf("}\n");
}

static void printExtenmoteN64Data(struct extenmote_n64_t *em_n64)
{
	printf("  N64 raw data { ");
	printf("    raw: ");
	printHex(em_n64->data, 4);
	printf("\n    Buttons: ");
	printN64Buttons(em_n64->data);
	printf("\n    Analog: x=0x%02x, y=0x%02x\n",
			em_n64->data[2],
			em_n64->data[3]);
	printf("  }\n");
}

static void printExtenmoteGCData(struct extenmote_gc_t *em_gc)
{
	printf("  Gamecube raw data {\n");
	printf("    raw: ");
	printHex(em_gc->data, 8);
	printf("\n    Buttons: ");
	printGCButtons(em_gc->data);
	printf("\n    Analog: x=0x%02x, y=0x%02x, cx=0x%02x, cy=0x%02x, l=0x%02x, r=0x%02x\n",
				em_gc->data[2],
				em_gc->data[3],
				em_gc->data[4],
				em_gc->data[5],
				em_gc->data[6],
				em_gc->data[7]);
	printf("  }\n");
}

static void printExtenmoteSNESData(struct extenmote_snes_t *em_snes)
{
	printf("  SNES raw data {\n");
	printf("    raw: ");
	printHex(em_snes->data, 2);
	printf("      Buttons: ");
	printSNESButtons(em_snes->data);
	printf("\n  }\n");
}

static void printExtenmoteNESData(struct extenmote_nes_t *em_nes)
{
	printf("  NES raw data {\n");
	printf("    raw: ");
	printHex(em_nes->data, 1);
	printf("      Buttons: ");
	printNESButtons(em_nes->data);
	printf("\n  }\n");
}

static void printNunchukData(struct nunchuk_t *nunchuk)
{
	printf("  Nunchuk data {\n");

	printf("    Joystick angle: %.1f mag: %.1f\n", nunchuk->js.ang, nunchuk->js.mag);
	printf("    Raw accel data: 0x%04x 0x%04x 0x%04x\n", nunchuk->accel.x, nunchuk->accel.y, nunchuk->accel.z);

	printf("}\n");
}

//---------------------------------------------------------------------------------
int main(int argc, char **argv) {
//---------------------------------------------------------------------------------

	// If enabled, the extenmote adapter will present a different expansion
	// ID depending on the connected controller. If disabled, a classic controller
	// is emulated (with extra extenmote specific data appended)
	wiiuse_enable_extenmote_quirk(0);

	// Initialise the video system
	VIDEO_Init();
	
	// This function initialises the attached controllers
	WPAD_Init();
	
	// Obtain the preferred video mode from the system
	// This will correspond to the settings in the Wii menu
	rmode = VIDEO_GetPreferredMode(NULL);

	// Allocate memory for the display in the uncached region
	xfb = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
	
	// Initialise the console, required for printf
	console_init(xfb,20,20,rmode->fbWidth,rmode->xfbHeight,rmode->fbWidth*VI_DISPLAY_PIX_SZ);
	
	// Set up the video registers with the chosen mode
	VIDEO_Configure(rmode);
	
	// Tell the video hardware where our display memory is
	VIDEO_SetNextFramebuffer(xfb);
	
	// Make the display visible
	VIDEO_SetBlack(FALSE);

	// Flush the video register changes to the hardware
	VIDEO_Flush();

	// Wait for Video setup to complete
	VIDEO_WaitVSync();
	if(rmode->viTVMode&VI_NON_INTERLACE) VIDEO_WaitVSync();

	while(1) {
		struct expansion_t exp;
	
		// Call WPAD_ScanPads each loop, this reads the latest controller states
		WPAD_ScanPads();

		// WPAD_ButtonsDown tells us which buttons were pressed in this loop
		// this is a "one shot" state which will not fire again until the button has been released
		u32 pressed = WPAD_ButtonsDown(0);

		// We return to the launcher application via exit
		if ( pressed & WPAD_BUTTON_HOME ) exit(0);

		WPAD_Expansion(0, &exp);

		VIDEO_WaitVSync();

		printf("\x1b[2J");		
		printf("\x1b[4;0H");
		printf("Raphnet extenmote displayer\n\n");

		printf("Expansion: %02x (%s)          \n", exp.type, extension_name(exp.type));

		switch (exp.type)
		{
			case EXP_CLASSIC:
				printClassicData(&exp.classic);
				break;
			
			case EXP_EXTENMOTE_N64:
				printExtenmoteN64Data(&exp.em_n64);
				break;

			case EXP_EXTENMOTE_GC:
				printExtenmoteGCData(&exp.em_gc);
				break;

			case EXP_EXTENMOTE_SNES:
				printExtenmoteSNESData(&exp.em_snes);
				break;

			case EXP_EXTENMOTE_NES:
				printExtenmoteNESData(&exp.em_nes);
				break;

			case EXP_NUNCHUK:
				printNunchukData(&exp.nunchuk);
				break;

			default:
				printf("\nExtension display not implemented.\n");
		}	
	}

	return 0;
}
